package com.ningdatech.pmapi.todocenter.manage;
import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.json.JSONUtil;
import com.alibaba.excel.EasyExcel;
import com.alibaba.fastjson.JSONObject;
import com.alibaba.xxpt.gateway.shared.api.request.OapiMessageWorkNotificationRequest;
import com.alibaba.xxpt.gateway.shared.api.response.OapiMessageWorkNotificationResponse;
import com.alibaba.xxpt.gateway.shared.client.http.ExecutableClient;
import com.alibaba.xxpt.gateway.shared.client.http.IntelligentGetClient;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.ningdatech.basic.enumeration.Status;
import com.ningdatech.basic.exception.BizException;
import com.ningdatech.basic.util.NdDateUtils;
import com.ningdatech.pmapi.common.constant.DingConstant;
import com.ningdatech.pmapi.common.constant.ProjectDeclareConstants;
import com.ningdatech.pmapi.common.util.ExcelDownUtil;
import com.ningdatech.pmapi.common.util.ExcelExportStyle;
import com.ningdatech.pmapi.todocenter.bean.entity.ProcessComment;
import com.ningdatech.pmapi.todocenter.bean.entity.ProgressNode;
import com.ningdatech.pmapi.todocenter.bean.vo.ProcessProgressDetailVo;
import com.ningdatech.pmapi.todocenter.constant.HisProInsEndActId;
import com.ningdatech.pmapi.todocenter.enums.ProcessHandlerEnum;
import com.ningdatech.pmapi.todocenter.enums.ProcessStatusEnum;
import com.ningdatech.pmapi.todocenter.extension.cmd.BackToHisApprovalNodeCmd;
import com.ningdatech.pmapi.todocenter.model.dto.req.ReqProcessHandlerDTO;
import com.ningdatech.pmapi.todocenter.model.dto.res.ResToBeProjectListExportDTO;
import com.ningdatech.pmapi.todocenter.zwdd.model.MessageContent;
import com.ningdatech.pmapi.todocenter.zwdd.model.MessageText;
import com.ningdatech.pmapi.user.entity.UserInfo;
import com.ningdatech.pmapi.user.service.IUserInfoService;
import com.ningdatech.pmapi.user.util.LoginUserUtil;
import com.wflow.bean.do_.UserDo;
import com.wflow.bean.entity.WflowCcTasks;
import com.wflow.bean.entity.WflowModelHistorys;
import com.wflow.bean.entity.WflowModels;
import com.wflow.exception.BusinessException;
import com.wflow.mapper.WflowCcTasksMapper;
import com.wflow.mapper.WflowModelHistorysMapper;
import com.wflow.service.OrgRepositoryService;
import com.wflow.workflow.bean.dto.ProcessInstanceOwnerDto;
import com.wflow.workflow.bean.process.OrgUser;
import com.wflow.workflow.bean.process.ProcessNode;
import com.wflow.workflow.bean.process.enums.ApprovalModeEnum;
import com.wflow.workflow.bean.process.enums.NodeTypeEnum;
import com.wflow.workflow.bean.process.form.Form;
import com.wflow.workflow.bean.process.props.ApprovalProps;
import com.wflow.workflow.bean.vo.ProcessProgressVo;
import com.wflow.workflow.bean.vo.ProcessTaskVo;
import com.wflow.workflow.config.WflowGlobalVarDef;
import com.wflow.workflow.service.*;
import com.wflow.workflow.service.FormService;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.assertj.core.util.Maps;
import org.flowable.bpmn.model.*;
import org.flowable.bpmn.model.Process;
import org.flowable.engine.*;
import org.flowable.engine.history.HistoricActivityInstance;
import org.flowable.engine.history.HistoricProcessInstance;
import org.flowable.engine.impl.util.ProcessDefinitionUtil;
import org.flowable.engine.repository.ProcessDefinition;
import org.flowable.engine.runtime.Execution;
import org.flowable.engine.runtime.ProcessInstance;
import org.flowable.task.api.Task;
import org.flowable.task.api.TaskQuery;

import org.flowable.variable.api.history.HistoricVariableInstance;
import org.springframework.stereotype.Component;
import com.ningdatech.basic.model.PageVo;

import com.ningdatech.pmapi.todocenter.model.dto.req.ReqToBeProcessedDTO;
import com.ningdatech.pmapi.todocenter.model.dto.res.ResToBeProcessedDTO;
import lombok.RequiredArgsConstructor;


import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

import static com.ningdatech.pmapi.todocenter.constant.WorkNotice.PASS_MSG_TEMPLATE;
import static com.ningdatech.pmapi.todocenter.constant.WorkNotice.PASS_MSG_TEMPLATE2;

/**
 * @author CMM
 * @since 2023/01/12 16:09
 */
@Component
@RequiredArgsConstructor
@Slf4j
public class TodoCenterManage {

    private final TaskService taskService;
    private final RepositoryService repositoryService;
    private final RuntimeService runtimeService;
    private final UserDeptOrLeaderService userDeptOrLeaderService;
    private final ProcessInstanceService processService;
    private final FormService formService;
    private final ManagementService managementService;
    private final HistoryService historyService;
    private final WflowModelHistorysMapper modelHistorysMapper;
    private final ProcessNodeCatchService nodeCatchService;
    private final OrgRepositoryService orgRepositoryService;
    private final ProcessTaskService processTaskService;
    private final WflowCcTasksMapper ccTasksMapper;
    private final ProcessModelService processModelService;
    private final IUserInfoService userInfoService;
    private  final ExecutableClient CLIENT;
    /**
     * 待办中心项目列表查询
     * @param param
     * @return com.ningdatech.basic.model.PageVo<com.ningdatech.pmapi.todocenter.model.dto.res.ResToBeProcessedDTO>
     * @author CMM
     * @since 2023/02/01 17:44
     */
    public PageVo<ResToBeProcessedDTO> queryProjectList(ReqToBeProcessedDTO param) {
        // 获取登录用户ID
        // long userId = LoginUserUtil.getUserId();
        int userId = 381496;
        // 获取入参分页信息
        Integer pageSize = param.getPageSize();
        Integer pageNumber = param.getPageNumber();

        TaskQuery taskQuery = taskService.createTaskQuery();
        taskQuery.active().taskCandidateOrAssigned(String.valueOf(userId)).orderByTaskCreateTime().desc();
        Page<ResToBeProcessedDTO> page = param.page();

        List<Task> taskList = taskQuery.listPage(pageSize * (pageNumber - 1), pageSize);
        Set<String> staterUsers = new HashSet<>();
        List<ResToBeProcessedDTO> resVos = taskList.stream().map(task -> {
            String nodeId = task.getTaskDefinitionKey();
            String instanceId = task.getProcessInstanceId();

            ProcessProgressVo instanceProgress = processService.getInstanceProgress(nodeId, instanceId);
            Map<String, Object> formData = instanceProgress.getFormData();
            String status = instanceProgress.getStatus();
            Boolean temporarySupplement = (Boolean) formData.get(ProjectDeclareConstants.BasicInformation.TEMPORARY_SUPPLEMENT);
            ResToBeProcessedDTO res = new ResToBeProcessedDTO();
            temporarySupplement = true;
            if (temporarySupplement.equals(param.getIsSupplement())) {
                ProcessDefinition processDef = repositoryService.createProcessDefinitionQuery()
                        .processDefinitionId(task.getProcessDefinitionId())
                        .singleResult();
                //从缓存取
                ProcessInstanceOwnerDto owner = runtimeService.getVariable(task.getExecutionId(), "owner", ProcessInstanceOwnerDto.class);
                ProcessInstance processInstance = runtimeService.createProcessInstanceQuery().processInstanceId(task.getProcessInstanceId()).singleResult();
                staterUsers.add(owner.getOwner());
                ProcessTaskVo processTaskVo = ProcessTaskVo.builder()
                        .taskId(task.getId())
                        .taskName(task.getName())
                        .taskDefKey(task.getTaskDefinitionKey())
                        .processDefId(task.getProcessDefinitionId())
                        .executionId(task.getExecutionId())
                        .nodeId(nodeId)
                        .deployId(processDef.getDeploymentId())
                        .processDefName(processDef.getName())
                        .version(processDef.getVersion())
                        .instanceId(instanceId)
                        .ownerId(processInstance.getStartUserId())
                        .ownerDeptId(owner.getOwnerDeptId())
                        .ownerDeptName(owner.getOwnerDeptName())
                        .createTime(processInstance.getStartTime())
                        .taskCreateTime(task.getCreateTime())
                        .build();
                res.setProcessTaskInfo(processTaskVo);
                String projectName = (String) formData.get(ProjectDeclareConstants.BasicInformation.PROJECT_NAME);
                res.setProjectName(projectName);
                res.setReportUnitId(owner.getOwnerDeptId());
                res.setReportUnitName(owner.getOwnerDeptName());
                res.setReportAmount((Integer) formData.get(ProjectDeclareConstants.FundDeclareInfo.DECLARE_AMOUNT));
                res.setBudgetYear((Integer) formData.get(ProjectDeclareConstants.BasicInformation.BUDGET_YEAR));
                res.setProcessStatus(ProcessStatusEnum.getCodeByDesc(status));
                res.setProcessStatusName(status);
                res.setProcessLaunchTime(NdDateUtils.date2LocalDateTime(processInstance.getStartTime()));
            }
            return res;
        }).filter(v -> StringUtils.isBlank(param.getProjectName()) || v.getProjectName().contains(param.getProjectName()))
          .filter(v -> StringUtils.isBlank(param.getReportUnitName()) || v.getReportUnitName().contains(param.getReportUnitName()))
          .filter(v -> Objects.isNull(param.getBudgetYear()) || v.getBudgetYear().equals(param.getBudgetYear()))
          .filter(v -> Objects.isNull(param.getProcessLaunchStartTime()) || v.getProcessLaunchTime().isBefore(param.getProcessLaunchStartTime()))
          .filter(v -> Objects.isNull(param.getProcessLaunchEndTime()) || v.getProcessLaunchTime().isAfter(param.getProcessLaunchEndTime()))
          .collect(Collectors.toList());
        if (CollectionUtil.isNotEmpty(staterUsers)) {
            Map<String, OrgUser> userMap = userDeptOrLeaderService.getUserMapByIds(staterUsers);
            page.setRecords(resVos.stream().peek(v -> v.getProcessTaskInfo().setOwner(userMap.get(v.getProcessTaskInfo().getOwnerId()))).collect(Collectors.toList()));
        }
        return PageVo.of(resVos,resVos.size());
    }
    /**
     * 待办中心列表导出
     * @param response
     * @param param
     * @return void
     * @author CMM
     * @since 2023/02/01 17:44
     */
    public void exportProjectList(HttpServletResponse response, ReqToBeProcessedDTO param) {
        PageVo<ResToBeProcessedDTO> page =
                queryProjectList(param);
        List<ResToBeProcessedDTO> collect = (List<ResToBeProcessedDTO>) page.getRecords();
        String fileName = null;
        if (param.getIsSupplement()){
            fileName = "增补项目列表";
        }else {
            fileName = "非增补项目列表";
        }
        ExcelDownUtil.setFileName(fileName,response);
        //数据导出处理函数
        try {
            EasyExcel.write(response.getOutputStream(), ResToBeProjectListExportDTO.class)
                    .autoCloseStream(false)
                    .registerWriteHandler(ExcelExportStyle.formalStyle())
                    .sheet(fileName)
                    .doWrite(collect);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }
    /**
     * 流程处理相关操作
     * @param param
     * @return void
     * @author CMM
     * @since 2023/02/01 17:43
     */
    public void handler(ReqProcessHandlerDTO param) {
        Long userId = LoginUserUtil.getUserId();
        Task task = taskService.createTaskQuery().taskId(param.getTaskId()).active().singleResult();
        HashMap<String, Object> formData = new HashMap<>(32);
        if (Objects.isNull(task)) {
            throw new BizException("任务不存在");
        }
        if (hasComment(param.getAuditInfo())) {
            taskService.addComment(param.getTaskId(), param.getInstanceId(), JSONObject.toJSONString(param.getAuditInfo()));
        }
        switch (param.getAction()) {
            // 通过
            case PASS:
                formService.updateInstanceFormData(param.getInstanceId(), formData);
                doPass(task, param);
            // 盖章并通过
            case SEAL_PASS:
                formService.updateInstanceFormData(param.getInstanceId(), formData);
                doSealPass(task, param);
            // 驳回
            case REJECT:
                formService.updateInstanceFormData(param.getInstanceId(), formData);
                doReject(task, param);
                break;
            // 退回
            case BACK:
                formService.updateInstanceFormData(param.getInstanceId(), formData);
                doBackTask(task, userId, param);
                break;
            // 撤回
            case WITHDRAW:
                doWithDrawProcess(task);
                break;
            default:
                throw new IllegalStateException("Unexpected value: " + param.getAction());
        }
    }

    /**
     * 审批任务:驳回
     * @param task   当前任务
     * @param param 参数
     */
    private void doReject(Task task, ReqProcessHandlerDTO param) {
        Map<String, Object> var = new HashMap<>(16);
        var.put("approve_" + task.getId(), param.getAction());
        // TODO 中止流程并使项目进入对应状态,给项目创建人、流程发起人发送浙政钉工作通知：【项目名称】的【流程名称】被驳回，请及时处理。
        //获取流程定义
        Process process = ProcessDefinitionUtil.getProcess(task.getProcessDefinitionId());
        String projectName = getProjectName(task);
        // 获取根节点即流程发起节点
        FlowNode rootNode = (FlowNode) process.getFlowElement("root", true);
        // TODO 中止流程并使项目进入对应状态,给项目创建人、流程发起人发送浙政钉工作通知：【项目名称】的【流程名称】被驳回，请及时处理。
        sendWorkNoticeToStartUser(task,projectName,rootNode);
        taskService.complete(param.getTaskId(), var);
    }
    /**
     * 获取最新版本的流程配置
     * @param task 当前任务
     * @return com.wflow.bean.entity.WflowModels
     * @author CMM
     * @since 2023/02/01 17:30
     */
    private WflowModels getLastWflowModels(Task task) {
        WflowModels wflowModels = processModelService.getOne(Wrappers.lambdaQuery(WflowModels.class)
                .eq(WflowModels::getProcessDefId, task.getProcessDefinitionId()));
        if (ObjectUtil.isNull(wflowModels)) {
            log.warn("流程{}不存在", wflowModels.getFormId());
            throw new BusinessException("不存在该表单");
        }
        return wflowModels;
    }

    /**
     * 审批任务:盖章并通过
     * @param task   当前任务
     * @param param 参数
     */
    private void doSealPass(Task task, ReqProcessHandlerDTO param) {
        Map<String, Object> var = new HashMap<>(16);
        var.put("approve_" + task.getId(), param.getAction());
        // TODO 判断项目申报单位级别，区县单位申报有上级主管单位意见栏，市级单位没有

        // TODO 市级单位：为大数据局；区县单位：为大数据中心(根据附件区分？)

        taskService.complete(param.getTaskId(), var);
    }

    /**
     * 审批任务:通过
     * @param task   当前任务
     * @param param 参数
     */
    private void doPass(Task task, ReqProcessHandlerDTO param) {

        String projectName = getProjectName(task);
        Map<String, Object> var = new HashMap<>(16);
        var.put("approve_" + task.getId(), param.getAction());
        //获取流程定义
        Process process = ProcessDefinitionUtil.getProcess(task.getProcessDefinitionId());
        // 获取当前节点
        FlowNode currentNode  = (FlowNode) process.getFlowElement(task.getTaskDefinitionKey(), true);
        // 获取流程下一个节点浙政钉用户ID
        String nextUserId = getNextUserId(currentNode);
        // 若有下一个审核人，向其发送浙政钉工作通知：标题：审核任务 内容：【单位名称】的【项目名称】需要您审核。
        if (Objects.nonNull(nextUserId)){
            UserInfo auditUserInfo = userInfoService.getById(Long.valueOf(nextUserId));
            // TODO 向其发送浙政钉工作通知
            String msg = String.format(PASS_MSG_TEMPLATE,null,projectName);
            sendWorkNotice(auditUserInfo,msg);
        }else {
            // 若没有，向发起人发送浙政钉工作通知：【项目名称】已通过【流程名称】，请及时开始下一步操作。
            // TODO 向其发送浙政钉工作通知 获取根节点的孩子节点(即发起人节点)，向其发送浙政钉工作通知
            // 获取根节点即流程发起节点
            FlowNode rootNode = (FlowNode) process.getFlowElement("root", true);
            sendWorkNoticeToStartUser(task, projectName, rootNode);
        }
        taskService.complete(param.getTaskId(), var);
    }

    private void sendWorkNoticeToStartUser(Task task, String projectName, FlowNode rootNode) {
        String startUserId = getRootUserId(rootNode);
        UserInfo startUserInfo = userInfoService.getById(Long.valueOf(startUserId));
        // 从历史表获取最新版本的流程
        WflowModels wflowModels = getLastWflowModels(task);
        String formName = wflowModels.getFormName();
        String msg = String.format(PASS_MSG_TEMPLATE2, projectName, formName);
        sendWorkNotice(startUserInfo, msg);
    }

    /**
     * 获取流程发起节点的浙政钉用户ID
     * @param rootNode 根节点
     * @return java.lang.String
     * @author CMM
     * @since 2023/02/02 21:47
     */
    private String getRootUserId(FlowNode rootNode) {
        String rootUserId = null;
        // 输出连线
        List<SequenceFlow> outgoingFlows = rootNode.getOutgoingFlows();
        // 遍历返回下一个节点信息
        for (SequenceFlow currentOutgoingFlow : outgoingFlows) {
            // 类型自己判断
            FlowElement targetFlowElement = currentOutgoingFlow.getTargetFlowElement();
            // TODO 若要会签需判断候选人
            // 发起事件
            if (targetFlowElement instanceof StartEvent) {
                UserTask userTask = (UserTask) targetFlowElement;
                rootUserId = userTask.getAssignee();
                break;
            }
        }
        return rootUserId;

    }

    /**
     * 获取当前节点的下一个节点的浙政钉用户ID
     * @param currentNode 当前节点
     * @return java.lang.String 下一个节点的浙政钉用户ID
     * @author CMM
     * @since 2023/02/02 16:49
     */
    private String getNextUserId(FlowNode currentNode) {
        String nextUserId = null;
        // 输出连线
        List<SequenceFlow> outgoingFlows = currentNode.getOutgoingFlows();
        // 遍历返回下一个节点信息
        for (SequenceFlow currentOutgoingFlow : outgoingFlows) {
            // 类型自己判断
            FlowElement targetFlowElement = currentOutgoingFlow.getTargetFlowElement();
            // TODO 若要会签需判断候选人
            // 用户任务
            if (targetFlowElement instanceof UserTask) {
                UserTask userTask = (UserTask) targetFlowElement;
                nextUserId = userTask.getAssignee();
                break;
            }
        }
        return nextUserId;
    }
    /**
     * 获取任务节点的申报项目名称
     * @param task 当前任务
     * @return java.lang.String 项目名称
     * @author CMM
     * @since 2023/02/01 17:27
     */
    private String getProjectName(Task task) {
        String nodeId = task.getTaskDefinitionKey();
        String instanceId = task.getProcessInstanceId();
        ProcessProgressVo instanceProgress = processService.getInstanceProgress(nodeId, instanceId);
        Map<String, Object> formData = instanceProgress.getFormData();
        String projectName = (String) formData.get(ProjectDeclareConstants.BasicInformation.PROJECT_NAME);
        return projectName;
    }
    /**
     * 判断处理操作是否含有审核意见
     * @param comment 审核意见
     * @return boolean
     * @author CMM
     * @since 2023/02/01 17:33
     */
    private boolean hasComment(ProcessComment comment) {
        return Objects.nonNull(comment) && (StrUtil.isNotBlank(comment.getText())
                || CollectionUtil.isNotEmpty(comment.getAttachments()));
    }
    /**
     * 发送浙政钉工作通知
     * @param auditUserInfo 审核人信息
     * @param msg 要发送的工作通知
     * @return void
     * @author CMM
     * @since 2023/02/01 17:32
     */
    private void sendWorkNotice(UserInfo auditUserInfo, String msg) {
        // TODO 获取浙政钉唯一标识
        String dingKey = null;
        IntelligentGetClient intelligentGetClient = CLIENT.newIntelligentGetClient(DingConstant.WORKING_NOTICE);
        OapiMessageWorkNotificationRequest request = new OapiMessageWorkNotificationRequest();
        //消息体(参考下文示例消息格式)
        MessageText messageText = new MessageText();
        messageText.setMsgType("text");
        MessageContent messageContent = new MessageContent();

        messageContent.setContent(msg);
        messageText.setText(messageContent);
        request.setMsg(JSONUtil.toJsonStr(messageText));
        // 构建唯一的消息ID
        // String bizMsgId = "ZB_URGE_NOTICE_"  + "_" + auditUserInfo.getDeptId() + "_" + auditUserInfo.getUserId();
        // request.setBizMsgId(bizMsgId);
        request.setBizMsgId(null);
        //租户id
        // request.setTenantId(GovDingProperties.tenantId.toString());
        request.setReceiverIds(dingKey);
        // 获取结果
        OapiMessageWorkNotificationResponse apiResult = intelligentGetClient.get(request);
        if (!apiResult.getSuccess() || !JSONUtil.parseObj(apiResult.getContent()).getBool("success")) {
            log.warn("发送工作通知失败: {}", apiResult.getContent());
            throw new BizException(Status.BAD_REQUEST.toString());
        }

    }
    /**
     * 撤销流程处理
     *
     * @param task 当前任务
     */
    private void doWithDrawProcess(Task task) {
        //获取流程定义
        Process process = ProcessDefinitionUtil.getProcess(task.getProcessDefinitionId());
        // 获取当前运行流程的发起人节点信息
        FlowNode rootNode = (FlowNode) process.getFlowElement("root", true);
        // 获取当前节点
        FlowNode currentNode = (FlowNode) process.getFlowElement(task.getTaskDefinitionKey(), true);
        UserTask userTask = (UserTask) process.getFlowElement(task.getTaskDefinitionKey());
        String currentUserId = userTask.getAssignee();
        String rootUserId = getRootUserId(rootNode);
        // 判断当前登录用户是否是流程发起人
        Long userId = LoginUserUtil.getUserId();
        List<Execution> executions = runtimeService.createExecutionQuery()
                .processInstanceId(task.getProcessInstanceId())
                .onlyChildExecutions().list();
        if (rootUserId.equals(String.valueOf(userId))){
            // TODO 若是流程发起人点击撤回，项目回到上一个状态，并删除当前审核人对应的待办记录
            // 强制流程指向撤回
            runtimeService.createChangeActivityStateBuilder()
                    .processInstanceId(task.getProcessInstanceId())
                    .moveActivityIdTo(task.getTaskDefinitionKey(), HisProInsEndActId.WITHDRAW)
                    .moveExecutionsToSingleActivityId(executions.stream().map(Execution::getId)
                            .collect(Collectors.toList()), HisProInsEndActId.WITHDRAW)
                    .changeState();
        }else {
            FlowElementsContainer parentContainer = currentNode.getParentContainer();
            for (FlowElement flowElement : parentContainer.getFlowElements()) {
                UserTask BeforeUserTask = (UserTask) flowElement;
                if (BeforeUserTask.getAssignee().equals(String.valueOf(userId))){
                    // TODO 若是前一个审核人点击撤回，在审核记录中移除自己提交过的审核意见、
                    //  待我处理中移除当前审核人的待办记录、
                    //  待我处理中增加自己的待办记录、
                    //  我已处理中去掉自己之前处理的记录
                    // 强制流程指向撤回
                    runtimeService.createChangeActivityStateBuilder()
                            .processInstanceId(task.getProcessInstanceId())
                            .moveActivityIdTo(flowElement.getId(), HisProInsEndActId.WITHDRAW)
                            .moveExecutionsToSingleActivityId(executions.stream().map(Execution::getId)
                                    .collect(Collectors.toList()), HisProInsEndActId.WITHDRAW)
                            .changeState();
                    break;
                }
            }
        }
    }

    /**
     * 退回流程处理
     *
     * @param task
     * @param task 当前任务
     * @param userId  当前登录用户ID
     * @param param   参数
     */
    private void doBackTask(Task task, Long userId, ReqProcessHandlerDTO param) {
        // 获取项目名称
        String projectName = getProjectName(task);
        //获取流程定义
        Process process = ProcessDefinitionUtil.getProcess(task.getProcessDefinitionId());
        // 获取根节点即流程发起节点
        FlowNode rootNode = (FlowNode) process.getFlowElement("root", true);
        // TODO 流程变成【被退回】状态，待我处理中，为流程发起人增加一条待办记录，给项目创建人、流程发起人发送浙政钉工作通知：【项目名称】的【流程名称】被退回，请及时处理。
        sendWorkNoticeToStartUser(task,projectName,rootNode);
        //执行自定义回退逻辑，回退到流程发起人
        managementService.executeCommand(new BackToHisApprovalNodeCmd(runtimeService, param.getTaskId(), rootNode.getId()));
        runtimeService.setVariables(param.getInstanceId(), Maps.newHashMap("approve_" + param.getTaskId(), param.getAction()));
        log.info("用户[{}] 退回流程[{}] [{} -> {}]", userId, param.getInstanceId(), task.getTaskDefinitionKey(), param.getTargetNode());
    }

    /**
     * 查询流程表单数据及审批的进度步骤
     * @param instanceId 流程实例ID
     * @param nodeId 当前获取流程人员关联的流程节点ID
     * @return 流程进度及表单详情
     */
    public ProcessProgressDetailVo getProcessDetail(String nodeId, String instanceId) {
        HistoricProcessInstance instance = historyService.createHistoricProcessInstanceQuery().processInstanceId(instanceId).singleResult();
        // 取表单及表单数据
        HistoricVariableInstance forms = historyService.createHistoricVariableInstanceQuery()
                .processInstanceId(instanceId).variableName(WflowGlobalVarDef.WFLOW_FORMS).singleResult();
        List<HistoricVariableInstance> formDatas = historyService.createHistoricVariableInstanceQuery()
                .processInstanceId(instanceId).variableNameLike("field%").list();
        // 取节点设置
        HistoricVariableInstance nodeProps = historyService.createHistoricVariableInstanceQuery()
                .processInstanceId(instanceId).variableName(WflowGlobalVarDef.WFLOW_NODE_PROPS).singleResult();
        Map<String, Object> nodePropsValue = (Map<String, Object>) nodeProps.getValue();
        ProcessNode<?> currentNode = null;
        if (StrUtil.isNotBlank(nodeId)) {
            // 搜索当前版本流程的配置
            WflowModelHistorys modelHistory = modelHistorysMapper.selectOne(new QueryWrapper<>(WflowModelHistorys.builder()
                    .processDefId(instance.getProcessDefinitionId()).version(instance.getProcessDefinitionVersion()).build()));
            currentNode = nodeCatchService.reloadProcessByStr(modelHistory.getProcess()).get(nodeId);
        }
        UserDo users = orgRepositoryService.getUserById(instance.getStartUserId());
        OrgUser startUser = OrgUser.builder().id(users.getUserId()).name(users.getUserName()).avatar(users.getAvatar()).build();
        List<ProgressNode> taskRecords = getHisTaskRecords(instanceId, nodePropsValue);
        // 获取添加抄送任务
        taskRecords.addAll(getCcTaskRecords(instanceId));
        if (ObjectUtil.isNull(instance.getEndTime())) {
            // TODO 下版实现 获取等待中且还未开始的任务，如果存在条件则需要直接解析条件
            taskRecords.addAll(getFutureTask(instanceId));
        }
        taskRecords = taskRecords.stream()
                .sorted(Comparator.comparing(ProgressNode::getStartTime))
                .collect(Collectors.toList());
        taskRecords.add(0, ProgressNode.builder()
                .nodeId("root")
                .name("提交申请")
                .user(startUser)
                .nodeType(NodeTypeEnum.ROOT)
                .startTime(instance.getStartTime())
                .finishTime(instance.getStartTime())
                .taskId("root")
                .result(ProcessHandlerEnum.PASS)
                .build());
        // 提取全量表单数据
        Map<String, Object> formData = formDatas.stream().collect(Collectors.toMap(HistoricVariableInstance::getVariableName, HistoricVariableInstance::getValue));
        HistoricVariableInstance variableInstance = historyService.createHistoricVariableInstanceQuery()
                .processInstanceId(instanceId).variableName("owner").singleResult();
        ProcessInstanceOwnerDto owner = (ProcessInstanceOwnerDto) variableInstance.getValue();
        ProcessProgressDetailVo res = ProcessProgressDetailVo.builder()
                .instanceId(instanceId)
                .version(instance.getProcessDefinitionVersion())
                .formItems(formService.filterFormAndDataByPermConfig((List<Form>) forms.getValue(), formData, currentNode))
                .formData(formData)
                .processDefName(instance.getProcessDefinitionName())
                .staterUser(startUser)
                .starterDept(null == owner ? null : owner.getOwnerDeptName())
                .result(instance.getEndActivityId())
                .startTime(instance.getStartTime())
                .progress(taskRecords)
                .build();
        if (Objects.isNull(instance.getEndActivityId())){
            res.setStatus(ProcessStatusEnum.UNDER_REVIEW.name());
        } else if (HisProInsEndActId.BACK.equals(instance.getEndActivityId())) {
            // TODO 被退回的审核节点状态这里只是暂时这么判断，具体怎么保存这个退回状态，后面讨论
            res.setStatus(ProcessStatusEnum.BE_BACKED.name());
        } else if (HisProInsEndActId.REJECT.equals(instance.getEndActivityId())) {
            res.setStatus(ProcessStatusEnum.BE_REJECTED.name());
        } else if (HisProInsEndActId.END.equals(instance.getEndActivityId())) {
            res.setStatus(ProcessStatusEnum.APPROVED.name());
        }
        return res;
    }

    /**
     * 获取抄送的流程实例信息
     *
     * @param instanceId 实例ID
     * @return 抄送我的流程
     */
    private List<ProgressNode> getCcTaskRecords(String instanceId) {
        Set<String> ccUsers = new HashSet<>();
        List<ProgressNode> ccList = ccTasksMapper.selectList(new QueryWrapper<WflowCcTasks>()
                .eq("instance_id", instanceId)).stream().map(task -> {
            ccUsers.add(task.getUserId());
            return ProgressNode.builder()
                    .nodeId(task.getNodeId())
                    .nodeType(NodeTypeEnum.CC)
                    .name(task.getNodeName())
                    .user(OrgUser.builder().id(task.getUserId()).build())
                    .startTime(task.getCreateTime())
                    .finishTime(task.getCreateTime())
                    .build();
        }).collect(Collectors.toList());
        if (CollectionUtil.isNotEmpty(ccUsers)) {
            Map<String, OrgUser> userMap = userDeptOrLeaderService.getUserMapByIds(ccUsers);
            ccList.stream().peek(v -> v.setUser(userMap.get(v.getUser().getId()))).collect(Collectors.toList());
        }
        return ccList;
    }

    /**
     * 获取流程的审批历史记录
     *
     * @param instanceId 审批实例ID
     * @param nodeProps  节点设置
     * @return 历史记录列表
     */
    private List<ProgressNode> getHisTaskRecords(String instanceId, Map<String, Object> nodeProps) {
        List<HistoricActivityInstance> list = historyService.createHistoricActivityInstanceQuery()
                .processInstanceId(instanceId).orderByHistoricActivityInstanceStartTime().asc().list();
        Set<String> userSet = new HashSet<>();
        //获取节点处理结果
        Map<String, ProcessHandlerEnum> varMap = historyService.createHistoricVariableInstanceQuery()
                .processInstanceId(instanceId).variableNameLike("approve_%").list().stream()
                .collect(Collectors.toMap(HistoricVariableInstance::getVariableName, v -> (ProcessHandlerEnum) v.getValue()));
        List<ProgressNode> progressNodes = list.stream().filter(his -> ObjectUtil.isNotNull(his.getTaskId())).map(his -> {
            Object props = nodeProps.get(his.getActivityId());
            ApprovalModeEnum approvalMode = null;
            if (props instanceof ApprovalProps) {
                approvalMode = ((ApprovalProps) props).getMode();
            }
            userSet.add(his.getAssignee());
            return ProgressNode.builder()
                    .nodeId(his.getActivityId())
                    .name(his.getActivityName())
                    .nodeType(NodeTypeEnum.APPROVAL)
                    .user(OrgUser.builder().id(his.getAssignee()).build())
                    .startTime(his.getStartTime())
                    .finishTime(his.getEndTime())
                    .taskId(his.getTaskId())
                    .approvalMode(approvalMode)
                    .result(varMap.get("approve_" + his.getTaskId()))
                    .build();
        }).collect(Collectors.toList());
        if (CollectionUtil.isNotEmpty(userSet)) {
            Map<String, OrgUser> map = userDeptOrLeaderService.getUserMapByIds(userSet);
            progressNodes.forEach(n -> n.setUser(map.get(n.getUser().getId())));
        }
        return progressNodes;
    }

    private List<ProgressNode> getFutureTask(String instanceId) {
        //根据流程遍历后续节点，期间要穿越后续包含并行网关和条件网关的节点
        return Collections.emptyList();
    }
}
